// Model.cpp
#include "model.h"
#include <iostream>
#include <fstream>
#include <GL/gl.h>
#include <stdexcept>

void Model::load(string filename)
{
	// Create input stream object
	ifstream infile(filename.c_str());
	if (! infile.good()) {
		throw std::runtime_error("Unable to load file " + filename);
	}
	// Read the number of points
	int numpoints;
	infile >> numpoints;
	// Resize the vector to hold them
	points.resize(numpoints);
	// Then read the points' coordinates.
	for (int i=0; i<numpoints; i++) {
		infile >> points[i].x >> points[i].y >> points[i].z;
		// points[i].y = - points[i].y; // Don't invert y.
	}
	// Same for polygons
	int numpolys;
	infile >> numpolys;
	polys.resize(numpolys); // resize vector as before
	// Read the polygons in
	for (int i=0; i<numpolys; i++) {
		// Read the vertex list
		int numvertices;
		infile >> numvertices;
		if (numvertices < 3) {
			std::cerr << "Strange, a polygon with less than three vertices " <<
				"Number " << i << std::endl;
			// throw std::runtime_error("Polygon with <3 verts");
		}
		polys[i].points.resize(numvertices);
		for (int j=0; j<numvertices; j++) {
			infile >> polys[i].points[j];
		}
		// Calculate the surface normal
		if (numvertices >= 3) {
			Vect3d firstpoint, secondpoint, thirdpoint;
			firstpoint = points[polys[i].points[0]];
			secondpoint = points[polys[i].points[1]];
			thirdpoint = points[polys[i].points[2]];
			polys[i].normal = (thirdpoint - firstpoint) * (secondpoint - firstpoint);
			// polys[i].normal =  (secondpoint - firstpoint) * (thirdpoint - firstpoint);
			// Normalise the normal (i.e. unit length)
			// Otherwise opengl lighting goes wrong.
			polys[i].normal.normalise();
		}
	}
	// Work out min/max z/y
	minz = 1000;
	maxz = -1000;
	miny = 1000;
	maxy = -1000;
	for (int i=0; i<numpoints; i++) {
		Vect3d & p = points[i];
		minz = std::min(p.z, minz);
		miny = std::min(p.y, miny);
		maxz = std::max(p.z, maxz);
		maxy = std::max(p.y, maxy);
	}
	middlez = (maxz + minz) / 2;
	textureScale = 1.0 / max(maxz - minz, maxy - miny);
	// The offset of our Y coordinate, so it goes from whatever .. 1.0
	textureYOffset = 1.0 - ((maxy - miny) * textureScale);

	// All done.
}

void Model::draw(bool texture) const
{
	// Go through the polys
	int numpolys = polys.size();
	for (int i=0; i<numpolys; i++)
	{
		int numpoints = polys[i].points.size();
		// skip weird polys.
		if (numpoints < 3) {
			continue;
		}
		glBegin(GL_POLYGON);
		glNormal3f(polys[i].normal.x,
			polys[i].normal.y,
			polys[i].normal.z);
		// Go backwards around the poly.
		for (int j = (numpoints - 1); j >=0; j--) {
			int vertex = polys[i].points[j];
			Vect3d p = points[vertex]; // Point from this model.
			
			if (texture) {
				// calc tex coords.
				Vect3d tx = points[vertex];
				// tx.z -= middlez; 
				tx.y -= miny;
				tx.z *= textureScale;
				tx.y *= textureScale;
				tx.y += textureYOffset;
				tx.z += 0.5;
				glTexCoord2f(tx.z, tx.y);
			}
			glVertex3f(p.x,
				p.y,
				p.z);
		}
		glEnd();
	}
}

void Model::drawTweened(bool texture, const Model & to, float distance) const
{
	if (points.size() > to.points.size()) {
		drawMeTweened(texture, to, distance);
	} else {
		to.drawMeTweened(texture, *this, 1.0 - distance);
	}
}

void Model::drawMeTweened(bool texture, const Model & to, float distance) const
{
	
	// Go through the polys
	int numpolys = polys.size();
	for (int i=0; i<numpolys; i++)
	{
		int numpoints = polys[i].points.size();
		// skip weird polys.
		if (numpoints < 3) {
			continue;
		}
		glBegin(GL_POLYGON);
		glNormal3f(polys[i].normal.x,
			polys[i].normal.y,
			polys[i].normal.z);
		// Go backwards around the poly.
		int othermax = to.points.size() - 1;
		for (int j = (numpoints - 1); j >=0; j--) {
			int vertex = polys[i].points[j];
			Vect3d p = points[vertex]; // Point from this model.
			// Get a point from the other model, the same one if poss.
			Vect3d q = (vertex > othermax) ? to.points[othermax] : to.points[vertex];
			
			// Weighted average of p and q.
			p.scalarprod(1.0 - distance);
			q.scalarprod(distance);
			Vect3d weighted = p + q;
			
			if (texture) {
				// calc tex coords.
				Vect3d tx = weighted;
				// tx.z -= middlez; 
				tx.y -= miny;
				tx.z *= textureScale;
				tx.y *= textureScale;
				tx.y += textureYOffset;
				tx.z += 0.5;
				glTexCoord2f(tx.z, tx.y);
			}
			glVertex3f(weighted.x,
				weighted.y,
				weighted.z);
		}
		glEnd();
	}
}
